/**
 * Modifications to be made:
 * 1. Two possible modes: free-size windowing and fixed-scale windowing
 * !!DONE!! 2. Add and remove the newly added window
 * !!DONE!! 3. Expand the width or height of the window by a constant number of pixels
 * !!DONE!! 4. Expand the dimension (both width and height) of window by a constant scale (e.g. 0.1)
 * !!DONE!! 5. Move the window to left, right, up, and down
 */
#include "stdafx.h"

#define GUIDE_NONE 0
#define GUIDE_SQUARE 1
#define GUIDE_RECTANGLE 2
#define GUIDE_CIRCLE 3
#define GUIDE_ELLIPSE 4

#ifdef WIN32
//#include "stdafx.h"
#include <io.h>
#else
#include <sys/types.h>
#include <dirent.h>
#endif

#include "cv.h"
#include "cvaux.h"
#include "highgui.h"
#include <stdio.h>
#include <time.h>
#include <string>
#include <fstream>
#include <sstream>
#include <vector>
#include <algorithm>

using namespace std;

struct DPrecRect {
	int x;
	int y;
	double width;
	double height;
};

IplImage* image  = 0;
IplImage* image2 = 0;
double roi_x0 = -1;  // (roi_x0, roi_y0) is the coordinate of first corner
double roi_y0 = -1;
double roi_x1 = -1;  // (roi_x1, roi_y1) is the coordinate of second corner
double roi_y1 = -1;
float scale_width = 5;
float scale_height = 5;
vector<DPrecRect>* objects;
char input_dir[256]   = "rawdata";
char outputname[256]  = "annotation.txt";
char* output_dir  = "cropped";
char raw_image_ext[16] = ".jpg";
char* configurationfile = "config.xml";
char window_name[200] = "Object Marker";
bool FIXED_SCALE = false;
bool ACCEPT_POINT = false;

bool SHOW_GUIDE = false;
int GUIDE_SHAPE = GUIDE_NONE; // 0
//CvRect guide_rect = cvRect(10,10,50,100);
CvRect guide_rect = cvRect(0,0,0,0);

vector<string> filenamelist;
vector<vector<DPrecRect>>	rectlist;

void printHelp();
void redrawImage();
void on_mouse(int, int, int, int, void*);
void saveObjects(FILE*, const char*);
int cropImages(const char*);
void writeAnnotateHeader(FILE* fname);
void writeAnnotateLine(FILE* fpt, string imgfullpath, vector<DPrecRect> rectvec);

bool change_flag;

void saveConfiguration(char* config_name)
{
	CvFileStorage* fs = cvOpenFileStorage(config_name,NULL,CV_STORAGE_WRITE);
	cvWriteString(fs, "image_folder", input_dir);
	cvWriteString(fs, "output_name", outputname);
	cvWriteString(fs, "image_ext", raw_image_ext);

	if(GUIDE_SHAPE >0){
		cvStartWriteStruct(fs, "guide", CV_NODE_MAP, NULL);
		cvWriteComment(fs, "1:SQUARE, 2:RECTANGLE, 3:CIRCLE, 4:ELLIPSE", 0);
		cvWriteInt(fs, "shape", GUIDE_SHAPE);
		cvStartWriteStruct(fs, "position", CV_NODE_MAP, NULL);
		cvWriteInt(fs, "x", guide_rect.x);
		cvWriteInt(fs, "y", guide_rect.y);
		cvWriteInt(fs, "width", guide_rect.width);
		cvWriteInt(fs, "height", guide_rect.height);
		cvEndWriteStruct(fs);
		cvWriteInt(fs,"display", SHOW_GUIDE ? 1 : 0);
		cvEndWriteStruct(fs);
	}

	cvReleaseFileStorage(&fs);
}

void loadConfiguration(char* config_name)
{
	const char* text_string;
	CvFileStorage* fs = cvOpenFileStorage(config_name,NULL,CV_STORAGE_READ);
	if(fs!=NULL){
		text_string = cvReadStringByName(fs, NULL, "image_folder");
		strcpy(input_dir, text_string);
		text_string = cvReadStringByName(fs, NULL, "output_name");
		strcpy(outputname, text_string);
		text_string = cvReadStringByName(fs, NULL, "image_ext");
		strcpy(raw_image_ext, text_string);

		CvFileNode* fn = cvGetFileNodeByName(fs, NULL, "guide");
		if(fn!=NULL){
			GUIDE_SHAPE = cvReadIntByName(fs, fn, "shape", 0);
			CvFileNode* fn2 = cvGetFileNodeByName(fs, fn, "position");
			guide_rect.x = cvReadIntByName(fs, fn2, "x");
			guide_rect.y = cvReadIntByName(fs, fn2, "y");
			guide_rect.width = cvReadIntByName(fs, fn2, "width");
			guide_rect.height = cvReadIntByName(fs, fn2, "height");
			if(GUIDE_SHAPE == GUIDE_CIRCLE || GUIDE_SHAPE == GUIDE_SQUARE){
				guide_rect.height = guide_rect.width;
			}
		}
		SHOW_GUIDE = cvReadIntByName(fs, fn, "display") > 0 ? true : false;
		cvReleaseFileStorage(&fs);
	}
}

void printHelp() {
	printf("\nObject Marker: \n\t 摜̃IuWFNgʒuL^邽߂̃c[\n");
	printf("\tGunawan Herman, April 2006\n");
	printf("\tAdapted from ObjectMarker.cpp by A.Florian\n");
	printf("\tModified and localized by Takuya MINAGAWA in 2010\n");
	printf("\n");
	printf("\t̃vO͓͉摜'%s'tH_̒T܂\n", input_dir);
	printf("\tIuWFNg̈ʒu̓t@C'%s'ɃeLXgo͂܂\n", outputname);
	printf("\n");
	printf("------------------------------------------------------------\n");
	printf("|  btn  |               function                           |\n");
	printf("|-------|--------------------------------------------------|\n");
	printf("|Enter  | ݂̃}[J[t@CɋLqA               |\n");
	printf("|       | ̉摜ǂݍ                               |\n");
	printf("|Space  | EnterƓ                                      |\n");
	printf("|BS     | ЂƂỎ摜ɖ߂                             |\n");
	printf("|ESC    | vOI                                 |\n");
	printf("|d      | ԐV}[J[                         |\n");
	printf("|r      | Ot[̃}[J[PĂяo               |\n");
	printf("|8      | ԐV}[J[1 px֓               |\n");
	printf("|9      | ԐV}[J[10px֓               |\n");
	printf("|2      | ԐV}[J[1 px֓               |\n");
	printf("|3      | ԐV}[J[10px֓               |\n");
	printf("|4      | ԐV}[J[1 px֓               |\n");
	printf("|5      | ԐV}[J[10px֓               |\n");
	printf("|6      | ԐV}[J[1 pxE֓               |\n");
	printf("|7      | ԐV}[J[10pxE֓               |\n");
	printf("|w      | ԐV}[J[̕1 pxg傷             |\n");
	printf("|W      | ԐV}[J[̕1 pxk             |\n");
	printf("|h      | ԐV}[J[̍1 pxg傷           |\n");
	printf("|H      | ԐV}[J[̍1 pxk           |\n");
	printf("|z      | ԐV}[J[̂傫2%%g傷         |\n");
	printf("|Z      | ԐV}[J[̂傫2%%k         |\n");
	printf("|m      | }[J[̏cŒ^Œ             |\n");
	printf("|s      | }[J[̏cw肷                       |\n");
	printf("|p      | }[J[`łȂ_             |\n");
	printf("|e      | t[ɂ}[J[Wʏo         |\n");
	printf("|g      | KCh̕\/\                              |\n");
	printf("|G      | KCh̐ݒ                                     |\n");
	printf("|f      | 摜̃tH_ύX                           |\n");
	printf("|o      | o̓t@CύX                               |\n");
	printf("|O      | o̓t@C𐬌`ĐVɍ쐬                 |\n");
	printf("|j      | wԍ̉摜փWv                         |\n");
	printf("|t      | ̃wv\                                 |\n");
	printf("------------------------------------------------------------\n");
	printf("\n");
	printf("mode = %s\n", FIXED_SCALE ? "FIXED_SCALE" : "FREE_SIZE");
	if(FIXED_SCALE) {
		printf("Scale constant: %f x %f\n", scale_width, scale_height);	
	}
	printf("ACCEPT_POINT = %s\n", ACCEPT_POINT ? "YES" : "NO");
}


void redrawImage() {
    //printf("redrawImage()\n");
	image2 = cvCloneImage(image);

	if(SHOW_GUIDE){
		if(guide_rect.width >0 && guide_rect.height >0){
			switch(GUIDE_SHAPE){
			case GUIDE_SQUARE:
			case GUIDE_RECTANGLE:
				cvRectangle(image2, cvPoint(guide_rect.x, guide_rect.y),
					cvPoint(guide_rect.x + guide_rect.width, guide_rect.y + guide_rect.height),
					CV_RGB(255,255,0), 1);
				break;
				
			case GUIDE_CIRCLE:
			case GUIDE_ELLIPSE:
				cvEllipse(image2, cvPoint((guide_rect.x + guide_rect.width)/2, (guide_rect.y + guide_rect.height)/2),
					cvSize(guide_rect.width / 2, guide_rect.height / 2), 0, 0, 360, CV_RGB(255,255,0), 1);
				break;

			}
		}
	}

	// Display all rectangles
	int numOfRect = (int)(objects->size());
	for(int i = 0; i < numOfRect; i++) {
		DPrecRect rectangle = objects->at(i);
		if(rectangle.width > 0 || rectangle.height > 0) {
			cvRectangle(image2, cvPoint(rectangle.x, rectangle.y), 
				cvPoint(rectangle.x + (int)(rectangle.width), rectangle.y + (int)(rectangle.height)),
				CV_RGB(255, 0, 0), 1);
		} else {
			cvCircle(image2, cvPoint(rectangle.x, rectangle.y), 1, CV_RGB(255, 0, 0));
		}
	}

	if(roi_x0 > 0) {
		cvRectangle(image2, cvPoint((int)roi_x0,(int)roi_y0), cvPoint((int)roi_x1,(int)roi_y1), CV_RGB(255,0,0),1);
	}

	cvShowImage(window_name, image2);
	cvReleaseImage(&image2);
}


void on_mouse(int event,int x,int y,int flag, void*) {

	// If left mouse button is pressed, record the first coordinate
	if(event == CV_EVENT_LBUTTONDOWN) {
	        //printf("Left mouse button pressed\n");
		roi_x0 = roi_x1 = x;
		roi_y0 = roi_y1 = y;
	}
  
	// If mouse is moved while left mouse button is still pressed
#ifdef WIN32
	if(event == CV_EVENT_MOUSEMOVE && flag == CV_EVENT_FLAG_LBUTTON) {
#else
	if(event == CV_EVENT_MOUSEMOVE && flag == 33) {
#endif
	        //printf("Move moved with left button pressed\n");
		roi_x1 = x;
		
		if(!FIXED_SCALE) {
			roi_y1 = y;
		} else {
			roi_y1 = roi_y0 + (((roi_x1 - roi_x0)/scale_width) * scale_height);
		}

		redrawImage();
	}

	// If left mouse button is released
	if(event == CV_EVENT_LBUTTONUP) {
	        //printf("Left button released\n");
		int topleft_x = min((int)roi_x0, (int)roi_x1);
		int topleft_y = min((int)roi_y0, (int)roi_y1);
		double width  = abs((int)roi_x0 - (int)roi_x1);
		double height = abs((int)roi_y0 - (int)roi_y1);

		if((width == 0 || height == 0) && !ACCEPT_POINT) return;

		DPrecRect rectangle = {topleft_x, topleft_y, width, height};
		objects->push_back(rectangle);
		roi_x0 = -1;  // indicates that there's no temporary rectangle
		redrawImage();
		change_flag = true;
	}
}


void saveObjects(FILE* output, const string imagename, bool empty_obj_save)
{
	int numOfRect = (int)(objects->size());
	if(numOfRect) {
		writeAnnotateLine(output, imagename, *objects);
		fflush(output);
	}
	else if(empty_obj_save){
		writeAnnotateLine(output, imagename, *objects);
		fflush(output);
	}
}

void saveObjects(FILE* output, const string imagename)
{
	saveObjects(output, imagename, false);
}

/*
int cropImages(const char* datafile) {
	static int index = 1;
	int counts = 0;

	FILE* in = fopen(datafile, "r");
	if(!in) {
		printf("Error: cannot open %s to read\n", datafile);
		exit(-1);
	}

	while(true) {
		char imagename[150], crop_output[20];
		int nrects, x, y, w, h;
		if(fscanf(in, "%s %d", imagename, &nrects) < 2) break;
		printf("%s %d\n", imagename, nrects);
		                                               
		IplImage* tmpimage = cvLoadImage(imagename, 0);
		IplImage* scaled_size = cvCloneImage(tmpimage);
		scaled_size->width = 24; //24;//16;
		scaled_size->height = 24; //12;//8;
		
		if(!tmpimage) {
			printf("Error: cannot load %s\n", imagename);
			exit(-1);
		}

		for(int i = 0; i < nrects; i++) {
			if(fscanf(in, "%d %d %d %d", &x, &y, &w, &h) < 4) {
				printf("%s %d %d\n", imagename, i, nrects);
				printf("Invalid structure in %s\n", datafile);
				exit(-1);
			}

			struct _IplROI roi = {0, x, y, w, h};
			tmpimage->roi = &roi;
			sprintf(crop_output, "%s/%05d.jpg", output_dir, index++);
			cvSaveImage(crop_output, tmpimage);
			
			// Load the object, and scale it to desired size
			IplImage* original_size = cvLoadImage(crop_output, 0);
			cvResize(original_size, scaled_size);
			// Overwrite the original object with scaled object
			cvSaveImage(crop_output, scaled_size);
			cvReleaseImage(&original_size);
			
			counts++;
			if(counts % 50 == 0) {
				printf("*");
			}
		}

		tmpimage->roi = NULL;
		cvReleaseImage(&tmpimage);
		cvReleaseImage(&scaled_size);
	}
	printf("\n");

	fclose(in);
	return counts;
}
*/

void writeAnnotateHeader(FILE* fname){
	time_t rawtime;
	struct tm * timeinfo;
	time(&rawtime);
	timeinfo = localtime(&rawtime);

	fprintf(fname, "\n");
	fprintf(fname, "########################\n");
	fprintf(fname, "# %s #", asctime(timeinfo));
	fprintf(fname, "########################\n");
}


void writeAnnotateLine(FILE* fpt, string imgfullpath, vector<DPrecRect> rectvec)
{

	int numOfRect = (int)(rectvec.size());
	fprintf(fpt, "%s %d ", imgfullpath.c_str(), numOfRect);
	for(int i = 0; i < numOfRect; i++) {
		DPrecRect rectangle = rectvec.at(i);
		fprintf(fpt, "%d %d %d %d ", rectangle.x, rectangle.y, (int)(rectangle.width), (int)(rectangle.height));
	}
	fprintf(fpt, "\n");
}


void saveTrimmedAnnotateFile(char* filename)
{
	FILE* trimf = fopen(filename, "w");
	if(!trimf) {
		printf("Error: cannot open %s to write to\n", outputname);
		return;
	}

	writeAnnotateHeader(trimf);
	int size = filenamelist.size();
	string imgfullpath;
	string dirpath = string(input_dir);
	if(dirpath == ""){
		dirpath = ".";
	}
	for(int i=0; i<size; i++){
		imgfullpath = dirpath + string("/") + filenamelist[i];
		writeAnnotateLine(trimf, imgfullpath, rectlist[i]);
	}
	fclose(trimf);
}


void clearRectList()
{
	vector<vector<DPrecRect>>::iterator rect_itr = rectlist.begin();
	while(rect_itr != rectlist.end()){
		rect_itr->clear();
		rect_itr++;
	}
}


bool checkRectPos(CvSize img_size, DPrecRect rect)
{
	if(rect.x >= img_size.width || rect.y >= img_size.height
		|| rect.x + rect.width < 0 || rect.y + rect.height < 0){
			return false;
	}
	else
	{
		return true;
	}
}


bool removeOutRangeRect()
{
	bool ret = false;

	vector<DPrecRect>::iterator rect_itr = objects->begin();
	while(rect_itr != objects->end()){
		if(!checkRectPos(cvGetSize(image), *rect_itr)){
			printf("A rectangle is out of image size.  Removed.\n");
			rect_itr = objects->erase(rect_itr);
			ret = true;
		}
		else{
			rect_itr++;
		}
	}
	return ret;
}


char* getStrBetweenSpace(char* str, int& num)
{
	int i=0;
	int b;
	bool flag_f=true;
	char int_str[50];
	char *ret_char;
	
	if(str[0]=='\0'){
		return NULL;
	}
	if(str[0]==' '){
		while(str[i]==' ' && str[i]!='\0'){
			i++;
		}
	}
	ret_char = str+i;
	b = i;

	while(str[i]!=' ' && str[i]!='\0'){
		i++;
	}
	num = i-b;

	return ret_char;
}


char* getRectStr(char* str, DPrecRect& rect){
	char* pp2 = str;
	int str_num;
	char numstr[50];

	pp2 = getStrBetweenSpace(pp2, str_num);
	if(pp2){
		strncpy(numstr, pp2, str_num);
		numstr[str_num]='\0';
		rect.x = atoi(numstr);
		pp2 += str_num;
	}
	else{
		return NULL;
	}

	pp2 = getStrBetweenSpace(pp2, str_num);
	if(pp2){
		strncpy(numstr, pp2, str_num);
		numstr[str_num]='\0';
		rect.y = atoi(numstr);
		pp2 += str_num;
	}
	else{
		return NULL;
	}

	pp2 = getStrBetweenSpace(pp2, str_num);
	if(pp2){
		strncpy(numstr, pp2, str_num);
		numstr[str_num]='\0';
		rect.width = atoi(numstr);
		pp2 += str_num;
	}
	else{
		return NULL;
	}

	pp2 = getStrBetweenSpace(pp2, str_num);
	if(pp2){
		strncpy(numstr, pp2, str_num);
		numstr[str_num]='\0';
		rect.height = atoi(numstr);
		pp2 += str_num;
	}
	else{
		return NULL;
	}

	return pp2;
}

string getFileNameFromPath(char* filefullpath)
{
	string::size_type	str_idx;
	string fullpathstr = filefullpath;
	str_idx = fullpathstr.find_last_of("\\");

	if(str_idx == string::npos){
		str_idx = fullpathstr.find_last_of("/");
	}

	if(str_idx == string::npos){
		str_idx = 0;
	}
	else{
		str_idx++;
	}

	return fullpathstr.substr(str_idx);
}

void loadAnnotateInfo()
{
	clearRectList();
	FILE* in = fopen(outputname, "r");
	if(!in) {
		return;
	}

	char imagename[150], strline[512];
	int ret;
	int str_num;
	int nrects;
	char *pp, *pp2;
	DPrecRect rect;
	vector<string>::iterator filelist_itr; 
	bool loop_f=true;
	int line = 0;

	while(loop_f) {
		pp = fgets(strline, 512, in);
		line ++;
		if(pp!=NULL && (ret=sscanf(strline, "%s %d", imagename, &nrects)) == 2){
			string fname = getFileNameFromPath(imagename);
//			printf("%s\n", fname.c_str());
			filelist_itr = find(filenamelist.begin(), filenamelist.end(), fname);
			if(filelist_itr!=filenamelist.end()){
				int idx = (int)distance(filenamelist.begin(), filelist_itr);

				pp2 = getStrBetweenSpace(strline, str_num);
				pp2 += str_num;
				pp2 = getStrBetweenSpace(pp2, str_num);
				pp2 += str_num;
				rectlist[idx].clear();
				for(int i=0; i<nrects; i++){
					if(pp2=getRectStr(pp2, rect)){
						rectlist[idx].push_back(rect);
					}
					else{
						printf("%s %d %d\n", imagename, i, nrects);
						printf("Line %d: Invalid structure in %s\n", line, outputname);
						getchar();
						exit(-1);
					}
				}
			}
		}
		else if(pp==NULL){
			loop_f = false;
		}
	}
}


#ifdef WIN32
void createFilenameList()
{
	// get *.bmp files in directory
	struct _finddata_t	bmp_file;
	long hFile;
	char pattern[150];
	char flag[8];
//	char tmpdirname[256];
//	sprintf(pattern, "%s/*%s", dirname.c_str(), raw_image_ext);
	sprintf(pattern, "%s/*%s", input_dir, raw_image_ext);
	while((hFile = (long)_findfirst(pattern, &bmp_file)) ==-1L) {
		printf("no appropriate input files in directory %s\n", input_dir);
		printf("Quit?(1:Yes, 0:No): ");
		scanf("%s", flag);
		if(!strcmp(flag, "1"))
			exit(0);
		printf("New Image Directory Name: ");
		scanf("%s", input_dir);
		saveConfiguration(configurationfile);
		sprintf(pattern, "%s/*%s", input_dir, raw_image_ext);
	}

	vector<DPrecRect> recvec;
	filenamelist.clear();
	rectlist.clear();

	do{
		filenamelist.push_back(bmp_file.name);
		rectlist.push_back(recvec);
	}while(_findnext(hFile, &bmp_file)==0);

	_findclose( hFile );
}
#endif


int main(int argc, char* argv[]) {
/*	if(argc > 1) {
		int n = cropImages(outputname);
		printf("%d images has been produced", n);
		if(n > 0) {
			printf(": %s/%05d.jpg - %s/%05d.jpg", output_dir, 1, output_dir, n);
		}
		printf("\n");
		exit(0);
	}
	
	for(int i = 0; i < argc; i++) {
		printf("%s\n", argv[i]);
	}*/

	bool reboot_flag;

	do{
	reboot_flag = false;	// Activate only when the image directory is changed.

	enum KeyBindings {
		Key_Enter = 13,  Key_ESC = 27,  Key_Space = 32, Key_d = 100, 
		Key_8 = 56, Key_9 = 57, Key_2 = 50, Key_3 = 51, Key_4 = 52, Key_5 = 53, Key_6 = 54, Key_7 = 55,
		Key_w = 119, Key_W = 87, Key_h = 104, Key_H = 72, Key_z = 122, Key_Z = 90, 
		Key_m = 109, Key_s = 115, Key_e = 101, Key_p = 112, Key_t = 116, Key_j = 106,
		Key_r = 114, Key_f = 102, Key_o = 111, Key_O = 79, Key_BS = 8, Key_G = 71, Key_g = 103
	};

	loadConfiguration(configurationfile);

	printHelp();

	createFilenameList();
	loadAnnotateInfo();
/*	vector<string>::iterator itr = filenamelist.begin();
	while(itr != filenamelist.end()){
		printf("%s\n",itr->c_str());
		itr++;
	}*/

//	struct _finddata_t bmp_file, bmp_file2;
//	long hFile, hFile2;
	int iKey = 0;
	FILE* output = fopen(outputname, "a");
	if(!output) {
		printf("Error: cannot open %s to write to\n", outputname);
		exit(-1);
	}

	writeAnnotateHeader(output);

	static int counter = 1;

	// get *.bmp files in directory
//	char pattern[150];
//	char flag[8];
//	char tmpdirname[256];

	// init highgui
	cvAddSearchPath(input_dir);
	string strPrefix;

	int img_list_size = filenamelist.size();
	string dirname = string(input_dir);
	if(dirname==""){
		dirname = ".";
	}

	for(int img_index=0; img_index < img_list_size; img_index++){
//		objects.clear();
		objects = &(rectlist[img_index]);
		strPrefix = dirname + string("/") + filenamelist[img_index];
   
		image = cvLoadImage(strPrefix.c_str(), 1);
		//window_name = bmp_file.name;
		sprintf(window_name, "%d - %s", img_index+1, filenamelist[img_index].c_str());
		cvNamedWindow(window_name, 1);
		cvSetMouseCallback(window_name, on_mouse);
//		cvShowImage(window_name, image);
		redrawImage();

		bool cont = false;
		change_flag = false;
		do {
			// Get user input
			iKey = cvWaitKey(0);

			switch(iKey) {
				// Press ESC to close this program, any unsaved changes will be discarded
				case Key_ESC:
					cvReleaseImage(&image);
					cvDestroyWindow(window_name);
					fclose(output);
					return 0;

				// Press Space or Enter to save marked objects on current image and proceed to the next image
				case Key_Space:
				case Key_Enter:
					cont = false;
					if(change_flag){
						if(removeOutRangeRect()){
							saveObjects(output, strPrefix, true);					
						}
						else{
							saveObjects(output, strPrefix, false);
						}
					}
					break;

				case Key_BS:
					if(img_index > 0){
						cont = false;
						img_index-=2;
					}
					else{
						cont = true;
					}
					break;

				// Press d to remove the last added object
				case Key_d:
					cont = true;
					if(!objects->empty()){
						change_flag = true;
						objects->pop_back();
						redrawImage();
						if(objects->empty()){
							saveObjects(output, strPrefix, true);
						}
					}
					break;

				case Key_8:
				case Key_9:
					cont = true;
					if(!objects->empty()) {
						change_flag = true;
						objects->back().y -= (iKey == Key_8) ? 1 : 10;
						redrawImage();
					}
					break;

				case Key_2:
				case Key_3:
					cont = true;
					if(!objects->empty()) {
						change_flag = true;
						objects->back().y += (iKey == Key_2) ? 1 : 10;
						redrawImage();
					}
					break;

				case Key_4:
				case Key_5:
					cont = true;
					if(!objects->empty()) {
						change_flag = true;
						objects->back().x -= (iKey == Key_4) ? 1 : 10;
						redrawImage();
					}
					break;

				case Key_6:
				case Key_7:
					cont = true;
					if(!objects->empty()) {
						change_flag = true;
						objects->back().x += (iKey == Key_6) ? 1 : 10;
						redrawImage();
					}
					break;

				case Key_w:
				case Key_W:
					cont = true;
					if(!objects->empty()) {
						change_flag = true;
						objects->back().width += (iKey == Key_w) ? 1 : ((objects->back().width > 5) ? -1 : 0);
						redrawImage();
					}
					break;

				case Key_h:
				case Key_H:
					cont = true;
					if(!objects->empty()) {
						change_flag = true;
						objects->back().height += (iKey == Key_h) ? 1 : ((objects->back().height > 5) ? -1 : 0);
						redrawImage();
					}
					break;

				case Key_z:
				case Key_Z:
					cont = true;
					if(!objects->empty()) {
						change_flag = true;
						objects->back().width *= (iKey == Key_z) ? 1.02 : ((objects->back().width > 5 && objects->back().height > 5) ? 0.98 : 1);
						objects->back().height *= (iKey == Key_z) ? 1.02 : ((objects->back().width > 5 && objects->back().height > 5) ? 0.98 : 1);
						redrawImage();
					}
					break;

				case Key_m:
					cont = true;
					FIXED_SCALE = !FIXED_SCALE;
					printf("mode = %s\n", FIXED_SCALE ? "FIXED_SCALE" : "FREE_SIZE");
					if(FIXED_SCALE) {
						if(!objects->empty()) {
							scale_width = objects->back().width;
							scale_height = objects->back().height;
						} else {
							scale_width = scale_height = 5;
						}

						printf("Scale constant: %f x %f\n", scale_width, scale_height);
					}
					break;

				case Key_s:
					cont = true;
					printf("Input the scale constant:\n");
					printf("\twidth : ");
					scanf("%f", &scale_width);
					printf("\theight: ");
					scanf("%f", &scale_height);
					printf("Scale constant: %f x %f\n", scale_width, scale_height);
					break;

				case Key_e:
					cont = true;
					for(int i = 0; i < (int)(objects->size()); i++) {
						printf("%d %d %f %f\n", objects->at(i).x, objects->at(i).y, objects->at(i).width, objects->at(i).height);	
					}
					break;

				case Key_p:
					cont = true;
					ACCEPT_POINT = !ACCEPT_POINT;
					printf("ACCEPT_POINT = %s\n", ACCEPT_POINT ? "YES" : "NO");
					break;

				case Key_t:
					cont = true;
					printHelp();
					break;

				case Key_j:
					int target_id;
					printf("Jump to image#: ");
					scanf("%d", &target_id);
					if(target_id > 0 && target_id <= img_list_size){
						img_index = target_id - 2;
						cont = false;
					}
					else{
						printf("image# must be between 1 and %d\n", img_list_size);
						cont = true;
					}
					break;

				case Key_r:
					cont = true;
					if(img_index > 0){
						if(!rectlist[img_index].empty() || !rectlist[img_index-1].empty()){
							change_flag = true;
							rectlist[img_index] = rectlist[img_index-1];
							redrawImage();
						}
					}
					break;

				case Key_g:
					cont = true;
					SHOW_GUIDE = !SHOW_GUIDE;
					printf("GUIDE: %s\n", SHOW_GUIDE ? "ON" : "OFF");
					redrawImage();
					saveConfiguration(configurationfile);
					break;

				case Key_G:
					cont = true;
					printf("Choose shape (0:NONE, 1:SQUARE, 2:RECTANGLE, 3:CIRCLE, 4:ELLIPSE): ");
					int shape_id;
					char num_char[8];
					scanf("%s", num_char);
					shape_id = atoi(num_char);
					if(shape_id > 0 && shape_id < 5){
						GUIDE_SHAPE = shape_id;
						printf("input x: ");
						scanf("%s", num_char);
						guide_rect.x = atoi(num_char);
						printf("input y: ");
						scanf("%s", num_char);
						guide_rect.y = atoi(num_char);
						printf("input width: ");
						scanf("%s", num_char);
						guide_rect.width = atoi(num_char);
						if(GUIDE_SHAPE == GUIDE_SQUARE || GUIDE_SHAPE == GUIDE_CIRCLE){
							guide_rect.height = guide_rect.width;
						}
						else{
							printf("input height: ");
							scanf("%s", num_char);
							guide_rect.height = atoi(num_char);
						}
						
						printf("position (x, y, width, height) = (%d, %d, %d, %d)\n", guide_rect.x, guide_rect.y, guide_rect.width, guide_rect.height);

						SHOW_GUIDE = true;
						printf("GUIDE: %s\n", SHOW_GUIDE ? "ON" : "OFF");
						redrawImage();
						saveConfiguration(configurationfile);
					}
					else if(shape_id==0){
						GUIDE_SHAPE = shape_id;
						SHOW_GUIDE = false;
						printf("GUIDE: OFF\n");
						redrawImage();
						saveConfiguration(configurationfile);
					}
					else{
						printf("Wrong Input\n");
					}
					break;

				case Key_f:
					cont = false;
					printf("New Image Directory Name: ");
					scanf("%s", input_dir);
					saveConfiguration(configurationfile);
					reboot_flag = true;
					img_index = img_list_size;
					break;

				case Key_o:
					cont = true;
					printf("New Output File Name: ");
					scanf("%s", outputname);
					saveConfiguration(configurationfile);
					fclose(output);
					loadAnnotateInfo();
					redrawImage();

					output = fopen(outputname, "a");
					if(!output) {
						printf("Error: cannot open %s to write to\n", outputname);
						exit(-1);
					}
					writeAnnotateHeader(output);
					break;

				case Key_O:
					cont = true;
					printf("Output File Name: ");
					char	trimfile[256];
					scanf("%s", trimfile);
					saveTrimmedAnnotateFile(trimfile);
					break;

				default:
					if(iKey >= 0 && iKey <= 127) {
						cont = true;
						printf("Unrecognised command\n");
					} else {
						cont = false;
					}
				}
			} while(cont);

			cvDestroyWindow(window_name);
			cvReleaseImage(&image);
		}
    
		fclose(output);
		}while(reboot_flag);
  
	return 0;
}

